import { OrderedSet, OrderedMap, Set } from "immutable";
import { TypeRef } from "./TypeBuilder";
import { TypeReconstituter, BaseGraphRewriteBuilder } from "./GraphRewriting";
import { TypeNames } from "./TypeNames";
import { TypeAttributes } from "./TypeAttributes";
export declare type PrimitiveStringTypeKind = "string" | "date" | "time" | "date-time";
export declare type PrimitiveTypeKind = "none" | "any" | "null" | "bool" | "integer" | "double" | PrimitiveStringTypeKind;
export declare type NamedTypeKind = "class" | "enum" | "union";
export declare type TypeKind = PrimitiveTypeKind | NamedTypeKind | "array" | "object" | "map" | "intersection";
export declare function isPrimitiveStringTypeKind(kind: TypeKind): kind is PrimitiveStringTypeKind;
export declare function isNumberTypeKind(kind: TypeKind): kind is "integer" | "double";
export declare function isPrimitiveTypeKind(kind: TypeKind): kind is PrimitiveTypeKind;
export declare abstract class Type {
    readonly typeRef: TypeRef;
    readonly kind: TypeKind;
    constructor(typeRef: TypeRef, kind: TypeKind);
    readonly abstract children: OrderedSet<Type>;
    directlyReachableTypes<T>(setForType: (t: Type) => OrderedSet<T> | null): OrderedSet<T>;
    getAttributes(): TypeAttributes;
    readonly hasNames: boolean;
    getNames(): TypeNames;
    getCombinedName(): string;
    readonly abstract isNullable: boolean;
    abstract isPrimitive(): this is PrimitiveType;
    abstract reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    readonly debugPrintKind: string;
    equals(other: any): boolean;
    hashCode(): number;
    protected abstract structuralEqualityStep(other: Type, queue: (a: Type, b: Type) => boolean): boolean;
    structurallyCompatible(other: Type): boolean;
    getParentTypes(): Set<Type>;
    getAncestorsNotInSet(set: Set<TypeRef>): Set<Type>;
}
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    constructor(typeRef: TypeRef, kind: PrimitiveTypeKind, checkKind?: boolean);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(_other: Type, _queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class StringType extends PrimitiveType {
    readonly enumCases: OrderedMap<string, number> | undefined;
    constructor(typeRef: TypeRef, enumCases: OrderedMap<string, number> | undefined);
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(_other: Type, _queue: (a: Type, b: Type) => boolean): boolean;
    readonly debugPrintKind: string;
}
export declare class ArrayType extends Type {
    private _itemsRef;
    readonly kind: "array";
    constructor(typeRef: TypeRef, _itemsRef?: TypeRef | undefined);
    setItems(itemsRef: TypeRef): undefined;
    private getItemsRef();
    readonly items: Type;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(other: ArrayType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class GenericClassProperty<T> {
    readonly typeData: T;
    readonly isOptional: boolean;
    constructor(typeData: T, isOptional: boolean);
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class ClassProperty extends GenericClassProperty<TypeRef> {
    constructor(typeRef: TypeRef, isOptional: boolean);
    readonly typeRef: TypeRef;
    readonly type: Type;
}
export declare class ObjectType extends Type {
    readonly isFixed: boolean;
    private _properties;
    private _additionalPropertiesRef;
    constructor(typeRef: TypeRef, kind: TypeKind, isFixed: boolean, _properties: OrderedMap<string, ClassProperty> | undefined, _additionalPropertiesRef: TypeRef | undefined);
    setProperties(properties: OrderedMap<string, ClassProperty>, additionalPropertiesRef: TypeRef | undefined): undefined;
    getProperties(): OrderedMap<string, ClassProperty>;
    getSortedProperties(): OrderedMap<string, ClassProperty>;
    getAdditionalProperties(): Type | undefined;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(other: ObjectType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class ClassType extends ObjectType {
    kind: "class";
    constructor(typeRef: TypeRef, isFixed: boolean, properties: OrderedMap<string, ClassProperty> | undefined);
}
export declare class MapType extends ObjectType {
    readonly kind: "map";
    constructor(typeRef: TypeRef, valuesRef: TypeRef | undefined);
    readonly values: Type;
}
export declare class EnumType extends Type {
    readonly cases: OrderedSet<string>;
    kind: "enum";
    constructor(typeRef: TypeRef, cases: OrderedSet<string>);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    isPrimitive(): this is PrimitiveType;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
    protected structuralEqualityStep(other: EnumType, _queue: (a: Type, b: Type) => void): boolean;
}
export declare function setOperationCasesEqual(ma: OrderedSet<Type>, mb: OrderedSet<Type>, membersEqual: (a: Type, b: Type) => boolean): boolean;
export declare abstract class SetOperationType extends Type {
    private _memberRefs;
    constructor(typeRef: TypeRef, kind: TypeKind, _memberRefs?: OrderedSet<TypeRef> | undefined);
    setMembers(memberRefs: OrderedSet<TypeRef>): void;
    protected getMemberRefs(): OrderedSet<TypeRef>;
    readonly members: OrderedSet<Type>;
    readonly sortedMembers: OrderedSet<Type>;
    readonly children: OrderedSet<Type>;
    isPrimitive(): this is PrimitiveType;
    protected structuralEqualityStep(other: SetOperationType, queue: (a: Type, b: Type) => boolean): boolean;
}
export declare class IntersectionType extends SetOperationType {
    kind: "intersection";
    constructor(typeRef: TypeRef, memberRefs?: OrderedSet<TypeRef>);
    readonly isNullable: boolean;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
}
export declare class UnionType extends SetOperationType {
    kind: "union";
    constructor(typeRef: TypeRef, memberRefs?: OrderedSet<TypeRef>);
    setMembers(memberRefs: OrderedSet<TypeRef>): void;
    readonly stringTypeMembers: OrderedSet<Type>;
    findMember(kind: TypeKind): Type | undefined;
    readonly isNullable: boolean;
    readonly isCanonical: boolean;
    reconstitute<T extends BaseGraphRewriteBuilder>(builder: TypeReconstituter<T>): void;
}
