"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeUtils_1 = require("./TypeUtils");
var TypeGraph_1 = require("./TypeGraph");
var TypeAttributes_1 = require("./TypeAttributes");
var Support_1 = require("./Support");
var TypeRef = /** @class */ (function () {
    function TypeRef(graph, index) {
        this.graph = graph;
        this.index = index;
    }
    TypeRef.prototype.deref = function () {
        return this.graph.atIndex(this.index);
    };
    TypeRef.prototype.equals = function (other) {
        if (!(other instanceof TypeRef)) {
            return false;
        }
        Support_1.assert(this.graph === other.graph, "Comparing type refs of different graphs");
        return this.index === other.index;
    };
    TypeRef.prototype.hashCode = function () {
        return this.index | 0;
    };
    return TypeRef;
}());
exports.TypeRef = TypeRef;
function provenanceToString(p) {
    return p
        .map(function (r) { return r.index; })
        .toList()
        .sort()
        .map(function (i) { return i.toString(); })
        .join(",");
}
// FIXME: Don't infer provenance.  All original types should be present in
// non-inferred form in the final graph.
exports.provenanceTypeAttributeKind = new TypeAttributes_1.TypeAttributeKind("provenance", Support_1.setUnion, function (a) { return a; }, provenanceToString);
exports.NoStringTypeMapping = {
    date: "date",
    time: "time",
    dateTime: "date-time"
};
var TypeBuilder = /** @class */ (function () {
    function TypeBuilder(_stringTypeMapping, alphabetizeProperties, _allPropertiesOptional, _addProvenanceAttributes, inheritsProvenanceAttributes) {
        var _this = this;
        this._stringTypeMapping = _stringTypeMapping;
        this.alphabetizeProperties = alphabetizeProperties;
        this._allPropertiesOptional = _allPropertiesOptional;
        this._addProvenanceAttributes = _addProvenanceAttributes;
        this.topLevels = immutable_1.Map();
        this.types = [];
        this.typeAttributes = [];
        this._addedForwardingIntersection = false;
        this.commitType = function (tref, t) {
            var index = tref.index;
            // const name = names !== undefined ? ` ${names.combinedName}` : "";
            // console.log(`committing ${t.kind}${name} to ${index}`);
            Support_1.assert(_this.types[index] === undefined, "A type index was committed twice");
            _this.types[index] = t;
        };
        // FIXME: make mutable?
        this._primitiveTypes = immutable_1.Map();
        this._noEnumStringType = undefined;
        // FIXME: Combine map and object type maps
        this._mapTypes = immutable_1.Map();
        this._arrayTypes = immutable_1.Map();
        this._enumTypes = immutable_1.Map();
        this._classTypes = immutable_1.Map();
        this._unionTypes = immutable_1.Map();
        this._intersectionTypes = immutable_1.Map();
        Support_1.assert(!_addProvenanceAttributes || !inheritsProvenanceAttributes, "We can't both inherit as well as add provenance");
        this.typeGraph = new TypeGraph_1.TypeGraph(this, _addProvenanceAttributes || inheritsProvenanceAttributes);
    }
    TypeBuilder.prototype.addTopLevel = function (name, tref) {
        // assert(t.typeGraph === this.typeGraph, "Adding top-level to wrong type graph");
        Support_1.assert(!this.topLevels.has(name), "Trying to add top-level with existing name");
        Support_1.assert(this.types[tref.index] !== undefined, "Trying to add a top-level type that doesn't exist (yet?)");
        this.topLevels = this.topLevels.set(name, tref);
    };
    TypeBuilder.prototype.reserveTypeRef = function () {
        var index = this.types.length;
        // console.log(`reserving ${index}`);
        this.types.push(undefined);
        var tref = new TypeRef(this.typeGraph, index);
        var attributes = this._addProvenanceAttributes
            ? exports.provenanceTypeAttributeKind.makeAttributes(immutable_1.Set([tref]))
            : immutable_1.Map();
        this.typeAttributes.push(attributes);
        return tref;
    };
    TypeBuilder.prototype.addType = function (forwardingRef, creator, attributes) {
        if (forwardingRef !== undefined) {
            Support_1.assert(this.types[forwardingRef.index] === undefined);
        }
        var tref = forwardingRef !== undefined ? forwardingRef : this.reserveTypeRef();
        if (attributes !== undefined) {
            this.addAttributes(tref, attributes);
        }
        var t = creator(tref);
        this.commitType(tref, t);
        return tref;
    };
    TypeBuilder.prototype.atIndex = function (index) {
        var maybeType = this.types[index];
        if (maybeType === undefined) {
            return Support_1.panic("Trying to deref an undefined type in a type builder");
        }
        var maybeNames = this.typeAttributes[index];
        return [maybeType, maybeNames];
    };
    TypeBuilder.prototype.addAttributes = function (tref, attributes) {
        if (attributes === undefined) {
            attributes = immutable_1.Map();
        }
        var index = tref.index;
        this.typeAttributes[index] = TypeAttributes_1.combineTypeAttributes(this.typeAttributes[index], attributes);
    };
    TypeBuilder.prototype.makeNullable = function (tref, attributes) {
        var t = Support_1.defined(this.types[tref.index]);
        if (t.kind === "null" || t.kind === "any") {
            return tref;
        }
        var nullType = this.getPrimitiveType("null");
        if (!(t instanceof Type_1.UnionType)) {
            return this.getUnionType(attributes, immutable_1.OrderedSet([tref, nullType]));
        }
        var _a = TypeUtils_1.removeNullFromUnion(t), maybeNull = _a[0], nonNulls = _a[1];
        if (maybeNull !== null)
            return tref;
        return this.getUnionType(attributes, nonNulls.map(function (nn) { return nn.typeRef; }).add(nullType));
    };
    TypeBuilder.prototype.finish = function () {
        this.typeGraph.freeze(this.topLevels, this.types.map(Support_1.defined), this.typeAttributes);
        return this.typeGraph;
    };
    TypeBuilder.prototype.addForwardingIntersection = function (forwardingRef, tref) {
        this._addedForwardingIntersection = true;
        return this.addType(forwardingRef, function (tr) { return new Type_1.IntersectionType(tr, immutable_1.OrderedSet([tref])); }, undefined);
    };
    TypeBuilder.prototype.forwardIfNecessary = function (forwardingRef, tref) {
        if (tref === undefined)
            return undefined;
        if (forwardingRef === undefined)
            return tref;
        return this.addForwardingIntersection(forwardingRef, tref);
    };
    Object.defineProperty(TypeBuilder.prototype, "didAddForwardingIntersection", {
        get: function () {
            return this._addedForwardingIntersection;
        },
        enumerable: true,
        configurable: true
    });
    TypeBuilder.prototype.getPrimitiveType = function (kind, forwardingRef) {
        Support_1.assert(kind !== "string", "Use getStringType to create strings");
        if (kind === "date")
            kind = this._stringTypeMapping.date;
        if (kind === "time")
            kind = this._stringTypeMapping.time;
        if (kind === "date-time")
            kind = this._stringTypeMapping.dateTime;
        if (kind === "string") {
            return this.getStringType(undefined, undefined, forwardingRef);
        }
        var tref = this.forwardIfNecessary(forwardingRef, this._primitiveTypes.get(kind));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.PrimitiveType(tr, kind); }, undefined);
            this._primitiveTypes = this._primitiveTypes.set(kind, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getStringType = function (attributes, cases, forwardingRef) {
        if (cases === undefined) {
            // FIXME: Right now we completely ignore names for strings
            // without enum cases.  That's the correct behavior at the time,
            // because string types never are assigned names, but we might
            // do that at some point, but in that case we'll want a different
            // type for each occurrence, not the same single string type with
            // all the names.
            //
            // The proper solution at that point might be to just figure
            // out whether we do want string types to have names (we most
            // likely don't), and if not, still don't keep track of them.
            var result = void 0;
            if (this._noEnumStringType === undefined) {
                result = this._noEnumStringType = this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, undefined); }, undefined);
            }
            else {
                result = this.forwardIfNecessary(forwardingRef, this._noEnumStringType);
            }
            this.addAttributes(this._noEnumStringType, attributes);
            return result;
        }
        return this.addType(forwardingRef, function (tr) { return new Type_1.StringType(tr, cases); }, attributes);
    };
    TypeBuilder.prototype.getEnumType = function (attributes, cases, forwardingRef) {
        var unorderedCases = cases.toSet();
        var tref = this.forwardIfNecessary(forwardingRef, this._enumTypes.get(unorderedCases));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.EnumType(tr, cases); }, attributes);
            this._enumTypes = this._enumTypes.set(unorderedCases, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueObjectType = function (attributes, properties, additionalProperties, forwardingRef) {
        var _this = this;
        properties = Support_1.mapOptional(function (p) { return _this.modifyPropertiesIfNecessary(p); }, properties);
        return this.addType(forwardingRef, function (tref) { return new Type_1.ObjectType(tref, "object", true, properties, additionalProperties); }, attributes);
    };
    TypeBuilder.prototype.getUniqueMapType = function (forwardingRef) {
        return this.addType(forwardingRef, function (tr) { return new Type_1.MapType(tr, undefined); }, undefined);
    };
    TypeBuilder.prototype.registerMapType = function (values, tref) {
        if (this._mapTypes.has(values))
            return;
        this._mapTypes = this._mapTypes.set(values, tref);
    };
    TypeBuilder.prototype.getMapType = function (values, forwardingRef) {
        var tref = this.forwardIfNecessary(forwardingRef, this._mapTypes.get(values));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.MapType(tr, values); }, undefined);
            this.registerMapType(values, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.registerClassType = function (properties, tref) {
        var map = properties.toMap();
        this._classTypes = this._classTypes.set(map, tref);
    };
    TypeBuilder.prototype.setObjectProperties = function (ref, properties, additionalProperties) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.ObjectType)) {
            return Support_1.panic("Tried to set properties of non-object type");
        }
        type.setProperties(this.modifyPropertiesIfNecessary(properties), additionalProperties);
        if (type instanceof Type_1.MapType) {
            this.registerMapType(Support_1.defined(additionalProperties), ref);
        }
        else if (type instanceof Type_1.ClassType) {
            if (!type.isFixed) {
                this.registerClassType(properties, ref);
            }
        }
    };
    TypeBuilder.prototype.getUniqueArrayType = function (forwardingRef) {
        return this.addType(forwardingRef, function (tr) { return new Type_1.ArrayType(tr, undefined); }, undefined);
    };
    TypeBuilder.prototype.registerArrayType = function (items, tref) {
        if (this._arrayTypes.has(items))
            return;
        this._arrayTypes = this._arrayTypes.set(items, tref);
    };
    TypeBuilder.prototype.getArrayType = function (items, forwardingRef) {
        var tref = this.forwardIfNecessary(forwardingRef, this._arrayTypes.get(items));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ArrayType(tr, items); }, undefined);
            this.registerArrayType(items, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.setArrayItems = function (ref, items) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.ArrayType)) {
            return Support_1.panic("Tried to set items of non-array type");
        }
        type.setItems(items);
        this.registerArrayType(items, ref);
    };
    TypeBuilder.prototype.modifyPropertiesIfNecessary = function (properties) {
        if (this.alphabetizeProperties) {
            properties = properties.sortBy(function (_, n) { return n; });
        }
        if (this._allPropertiesOptional) {
            properties = properties.map(function (cp) { return new Type_1.ClassProperty(cp.typeRef, true); });
        }
        return properties;
    };
    TypeBuilder.prototype.getClassType = function (attributes, properties, forwardingRef) {
        properties = this.modifyPropertiesIfNecessary(properties);
        var tref = this.forwardIfNecessary(forwardingRef, this._classTypes.get(properties.toMap()));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.ClassType(tr, false, properties); }, attributes);
            this.registerClassType(properties, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.registerUnionType = function (members, tref) {
        if (this._unionTypes.has(members))
            return;
        this._unionTypes = this._unionTypes.set(members, tref);
    };
    // FIXME: Maybe just distinguish between this and `getClassType`
    // via a flag?  That would make `ClassType.map` simpler.
    TypeBuilder.prototype.getUniqueClassType = function (attributes, isFixed, properties, forwardingRef) {
        var _this = this;
        properties = Support_1.mapOptional(function (p) { return _this.modifyPropertiesIfNecessary(p); }, properties);
        return this.addType(forwardingRef, function (tref) { return new Type_1.ClassType(tref, isFixed, properties); }, attributes);
    };
    TypeBuilder.prototype.getUnionType = function (attributes, members, forwardingRef) {
        var unorderedMembers = members.toSet();
        var tref = this.forwardIfNecessary(forwardingRef, this._unionTypes.get(unorderedMembers));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.UnionType(tr, members); }, attributes);
            this.registerUnionType(unorderedMembers, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueUnionType = function (attributes, members, forwardingRef) {
        return this.addType(forwardingRef, function (tref) { return new Type_1.UnionType(tref, members); }, attributes);
    };
    TypeBuilder.prototype.registerIntersectionType = function (members, tref) {
        if (this._intersectionTypes.has(members))
            return;
        this._intersectionTypes = this._intersectionTypes.set(members, tref);
    };
    TypeBuilder.prototype.getIntersectionType = function (attributes, members, forwardingRef) {
        var unorderedMembers = members.toSet();
        var tref = this.forwardIfNecessary(forwardingRef, this._intersectionTypes.get(unorderedMembers));
        if (tref === undefined) {
            tref = this.addType(forwardingRef, function (tr) { return new Type_1.IntersectionType(tr, members); }, attributes);
            this.registerIntersectionType(unorderedMembers, tref);
        }
        else {
            this.addAttributes(tref, attributes);
        }
        return tref;
    };
    TypeBuilder.prototype.getUniqueIntersectionType = function (attributes, members, forwardingRef) {
        return this.addType(forwardingRef, function (tref) { return new Type_1.IntersectionType(tref, members); }, attributes);
    };
    TypeBuilder.prototype.setSetOperationMembers = function (ref, members) {
        var type = ref.deref()[0];
        if (!(type instanceof Type_1.UnionType || type instanceof Type_1.IntersectionType)) {
            return Support_1.panic("Tried to set members of non-set-operation type");
        }
        type.setMembers(members);
        var unorderedMembers = members.toSet();
        if (type instanceof Type_1.UnionType) {
            this.registerUnionType(unorderedMembers, ref);
        }
        else if (type instanceof Type_1.IntersectionType) {
            this.registerIntersectionType(unorderedMembers, ref);
        }
        else {
            return Support_1.panic("Unknown set operation type ${type.kind}");
        }
    };
    TypeBuilder.prototype.setLostTypeAttributes = function () {
        return;
    };
    return TypeBuilder;
}());
exports.TypeBuilder = TypeBuilder;
